# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Web Application Firewall Binding
"""
import json
from ctypes import POINTER, Structure, byref, c_bool, c_char_p, c_int, \
    c_size_t, c_void_p

from . import get_lib
from ._compat import UNICODE_CLASS
from .input import C_PWArgs, PWArgs

PW_ERR_INTERNAL = -6
PW_ERR_TIMEOUT = -5
PW_ERR_INVALID_CALL = -4
PW_ERR_INVALID_RULE = -3
PW_ERR_INVALID_FLOW = -2
PW_ERR_NORULE = -1
PW_GOOD = 0
PW_MONITOR = 1
PW_BLOCK = 2


class C_PWRet(Structure):

    _fields_ = [
        ("action", c_int),
        ("data", c_char_p),
    ]


_waf_lib = None


def get_waf_lib():
    global _waf_lib

    if _waf_lib is None:
        lib = get_lib()
        lib.powerwaf_initWithDiag.argstype = [
            c_char_p, c_char_p, c_void_p, POINTER(c_char_p)]
        lib.powerwaf_initWithDiag.restype = c_bool
        lib.powerwaf_freeDiagnotics.argstype = [c_char_p]
        lib.powerwaf_freeDiagnotics.restype = None
        lib.powerwaf_run.argstype = [
            c_char_p, POINTER(C_PWArgs), c_size_t]
        lib.powerwaf_run.restype = POINTER(C_PWRet)
        lib.powerwaf_freeReturn.argstype = [POINTER(C_PWRet)]
        lib.powerwaf_freeReturn.restype = None
        lib.powerwaf_clearRule.argstype = [c_char_p]
        lib.powerwaf_clearRule.restype = None
        _waf_lib = lib
    return _waf_lib


def initialize_with_diag(rule_name, rule_data):
    """ Initialize a WAF rule with diagnostic.
    """

    if isinstance(rule_name, UNICODE_CLASS):
        rule_name = rule_name.encode("utf-8")

    if isinstance(rule_data, UNICODE_CLASS):
        rule_data = rule_data.encode("utf-8")

    if rule_name is not None and not isinstance(rule_name, bytes):
        raise ValueError("rule_name must be a string, bytes or None")

    if not isinstance(rule_data, bytes):
        raise ValueError("rule_data must be a string or bytes")

    diag = None
    diag_ptr = c_char_p(None)
    try:
        ret = get_waf_lib().powerwaf_initWithDiag(
            rule_name, rule_data, None, byref(diag_ptr))
        if diag_ptr:
            diag = json.loads(diag_ptr.value.decode("utf-8"))
        return ret, diag
    finally:
        get_waf_lib().powerwaf_freeDiagnotics(diag_ptr)


def initialize(rule_name, rule_data):
    """ Initialize a WAF rule.
    """
    return initialize_with_diag(rule_name, rule_data)[0]


def validate(rule_data):
    """ Ask PowerWAF to parse and validate a ruleset.
    """
    return initialize_with_diag(None, rule_data)[1]


def clear(rule_name):
    """ Clear a WAF rule.
    """
    if isinstance(rule_name, UNICODE_CLASS):
        rule_name = rule_name.encode("utf-8")

    if not isinstance(rule_name, bytes):
        raise ValueError("rule_name must be a string or bytes")

    get_waf_lib().powerwaf_clearRule(rule_name)


def get_version():
    """ Get the WAF runtime version.
    """
    return get_waf_lib().powerwaf_getVersion().as_tuple()


def run(rule_name, parameters, budget):
    """ Run a WAF rule.
    """
    if isinstance(rule_name, UNICODE_CLASS):
        rule_name = rule_name.encode("utf-8")

    if not isinstance(rule_name, bytes):
        raise ValueError("rule_name must be a string or bytes")

    if not isinstance(parameters, PWArgs):
        parameters = PWArgs.from_python(parameters)

    return PWRet(get_waf_lib().powerwaf_run(
        rule_name, byref(parameters._obj), c_size_t(budget)))


def free(result):
    """ Free the result of the run function.
    """
    get_waf_lib().powerwaf_freeReturn(result)


class PWRet:
    """
    Higher-level WAF return value.
    """

    def __init__(self, obj):
        self._obj = obj

    def __del__(self):
        if self._obj is None:
            return
        free(self._obj)
        self._obj = None

    @property
    def action(self):
        if self._obj is not None and self._obj[0]:
            return self._obj[0].action

    @property
    def data(self):
        if self._obj is not None and self._obj[0]:
            return self._obj[0].data

    def __repr__(self):
        return "<PWRet action={0.action!r} data={0.data!r}>".format(self)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019, 2020, Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

__author__ = "Sqreen"
__copyright__ = "Copyright 2016, 2017, 2018, 2019, 2020, Sqreen"
__email__ = "contact@sqreen.io"
__license__ = "proprietary"
# The lib version is always contained at the beginning of __version__
__version__ = "0.5.1.0.1"
__lib_version__ = "0.5.1"



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Sqreen Python Native Module"""
import ctypes
import logging
import os
import sys

try:
    import pkg_resources
except ImportError:
    pkg_resources = None  # pragma: no cover

from .__about__ import __lib_version__

LOGGER = logging.getLogger("sqreen.native")


class C_PWVersion(ctypes.Structure):

    _fields_ = [
        ("major", ctypes.c_uint16),
        ("minor", ctypes.c_uint16),
        ("patch", ctypes.c_uint16)
    ]

    def as_tuple(self):
        """Return the version as a tuple."""
        return int(self.major), int(self.minor), int(self.patch)

    def __str__(self):
        """Return the version as a string."""
        return ".".join([str(x) for x in self.as_tuple()])


def _get_lib_path(name):
    """Return the path of the library called `name`."""
    if os.name == "posix" and sys.platform == "darwin":
        prefix, ext = "lib", ".dylib"
    elif sys.platform == "win32":
        prefix, ext = "", ".dll"
    else:
        prefix, ext = "lib", ".so"
    fn = None
    meipass = getattr(sys, "_MEIPASS", None)
    if meipass is not None:
        fn = os.path.join(meipass, prefix + name + ext)
    if fn is None and pkg_resources is not None:
        fn = pkg_resources.resource_filename("sq_native", prefix + name + ext)
    if fn is None:
        root_dir = os.path.dirname(os.path.abspath(__file__))
        fn = os.path.join(root_dir, prefix + name + ext)
    return fn


def _load_library(name):
    """Load a native library located in this module."""
    path = _get_lib_path(name)
    LOGGER.debug("Loading %s", path)
    return ctypes.cdll.LoadLibrary(path)


LIB_LOG_LEVELS = [
    None,
    logging.DEBUG,
    logging.INFO,
    logging.WARNING,
    logging.ERROR
]

LOG_LEVEL_MAP = {
    "TRACE": 0,
    "DEBUG": 1,
    "INFO": 2,
    "WARNING": 3,
    "ERROR": 4,
}


LOG_FUNC_TYPE = ctypes.CFUNCTYPE(
    None, ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p,
    ctypes.c_int, ctypes.c_char_p, ctypes.c_size_t)


@LOG_FUNC_TYPE
def _log_callback(level, func, filename, line, msg, msg_len):
    try:
        level = LIB_LOG_LEVELS[level]
        if level is not None:
            LOGGER.log(level, "%s:%d:%s %s", filename.decode("utf-8"), line,
                       func.decode("utf-8"), msg.decode("utf-8"))
    except Exception:
        LOGGER.warning("Log callback from the native library has failed")


NOT_LOADED = object()
lib = NOT_LOADED


def get_lib():
    """Get an instance of the Sqreen library."""
    global lib

    if lib is None or lib is NOT_LOADED:
        lib = _load_library("Sqreen")
        lib.powerwaf_getVersion.argstype = []
        lib.powerwaf_getVersion.restype = C_PWVersion

        lib_version = str(lib.powerwaf_getVersion())
        if lib_version != __lib_version__:
            raise RuntimeError("Native library version mismatch: {} != {}"
                               .format(lib_version, __lib_version__))

        lib.powerwaf_setupLogging.argstype = [LOG_FUNC_TYPE, ctypes.c_int]
        lib.powerwaf_setupLogging.restype = ctypes.c_bool
        level = LOG_LEVEL_MAP.get(os.getenv("SQREEN_WAF_LOG_LEVEL", None))
        if level is not None:
            if not lib.powerwaf_setupLogging(_log_callback, level):
                raise RuntimeError("Native library logging error")

    return lib


def get_lib_version():
    """Return the Sqreen library version or None if absent."""
    path = _get_lib_path("Sqreen")
    if path is not None and os.path.exists(path):
        return __lib_version__
    return None


__all__ = ["get_lib", "get_lib_version"]



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Compatibility stub
"""
import sys

PYTHON_VERSION = sys.version_info[0]


if PYTHON_VERSION == 2:
    ALL_STRING_CLASS = basestring  # noqa
    STRING_CLASS = str
    UNICODE_CLASS = unicode  # noqa
elif PYTHON_VERSION == 3:
    ALL_STRING_CLASS = str
    STRING_CLASS = str
    UNICODE_CLASS = str



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Binding for the WAF Input Data Structure
"""
from ctypes import (POINTER, Structure, byref, c_bool, c_char_p, c_int,
                    c_int64, c_size_t, c_uint64, c_void_p)

from . import get_lib
from ._compat import UNICODE_CLASS

PWI_INVALID = 0
PWI_SIGNED_NUMBER = 1 << 0
PWI_UNSIGNED_NUMBER = 1 << 1
PWI_STRING = 1 << 2
PWI_ARRAY = 1 << 3
PWI_MAP = 1 << 4


class C_PWArgs(Structure):

    _fields_ = [
        ("name", c_char_p),
        ("name_length", c_uint64),
        ("value", c_void_p),
        ("nb_entries", c_uint64),
        ("type", c_int),
    ]


_input_lib = None


def get_input_lib():
    global _input_lib

    if _input_lib is None:
        lib = get_lib()
        lib.powerwaf_createArray.argstype = []
        lib.powerwaf_createArray.restype = C_PWArgs
        lib.powerwaf_createInt.argstype = [c_int64]
        lib.powerwaf_createInt.restype = C_PWArgs
        lib.powerwaf_createUint.argstype = [c_uint64]
        lib.powerwaf_createUint.restype = C_PWArgs
        lib.powerwaf_createMap.restype = C_PWArgs
        lib.powerwaf_createStringWithLength.argstype = [c_char_p, c_size_t]
        lib.powerwaf_createStringWithLength.restype = C_PWArgs

        lib.powerwaf_addToPWArgsArray.argtypes = [
                POINTER(C_PWArgs), C_PWArgs]
        lib.powerwaf_addToPWArgsArray.restype = c_bool
        lib.powerwaf_addToPWArgsMap.argtypes = [
            POINTER(C_PWArgs), c_char_p, c_size_t, C_PWArgs]
        lib.powerwaf_addToPWArgsMap.restype = c_bool

        lib.powerwaf_freeInput.argtypes = [POINTER(C_PWArgs), c_bool]
        lib.powerwaf_freeInput.restype = None
        _input_lib = lib
    return _input_lib


def create_array():
    return get_input_lib().powerwaf_createArray()


def create_int(value):
    return get_input_lib().powerwaf_createInt(c_int64(value))


def create_uint(value):
    return get_input_lib().powerwaf_createUint(c_uint64(value))


def create_map():
    return get_input_lib().powerwaf_createMap()


def create_string(value, max_string_length=4096):
    if isinstance(value, UNICODE_CLASS):
        value = value[:max_string_length].encode("utf-8")

    if not isinstance(value, bytes):
        raise ValueError("value must be a string or bytes")

    value = value[:max_string_length]
    return get_input_lib().powerwaf_createStringWithLength(value, len(value))


def append_to_array(array, value):
    return get_input_lib().powerwaf_addToPWArgsArray(byref(array), value)


def append_to_map(array, key, value):
    if isinstance(key, UNICODE_CLASS):
        key = key.encode("utf-8")

    if not isinstance(key, bytes):
        raise ValueError("value must be a string or bytes")

    return get_input_lib().powerwaf_addToPWArgsMap(byref(array), key, 0, value)


def free(value):
    get_input_lib().powerwaf_freeInput(byref(value), False)


def create(value, max_depth=10, ignore_none=True, max_string_length=4096,
           max_items=150):
    """ Lower-level function to convert a Python value to input value
    """
    if isinstance(value, str) or isinstance(value, bytes):
        return create_string(value, max_string_length=max_string_length)

    if isinstance(value, bool):
        return create_uint(int(value))

    if isinstance(value, int):
        if value < 0:
            return create_int(value)
        else:
            return create_uint(value)

    if isinstance(value, list) or isinstance(value, tuple):
        obj = create_array()
        if max_depth <= 0:
            # ignore if deeply nested
            return obj
        for i, item in enumerate(value):
            if i >= max_items or (item is None and ignore_none):
                continue
            item_obj = create(item, max_depth=max_depth - 1)
            ret = append_to_array(obj, item_obj)
            if ret is False:
                free(item_obj)
        return obj

    if isinstance(value, dict):
        obj = create_map()
        if max_depth <= 0:
            # ignore if deeply nested
            return obj
        for i, (k, v) in enumerate(value.items()):
            if i >= max_items or (v is None and ignore_none):
                continue
            item_obj = create(v, max_depth=max_depth - 1)
            ret = append_to_map(obj, k, item_obj)
            if ret is False:
                free(item_obj)
        return obj

    return create_string(UNICODE_CLASS(value), max_string_length=max_string_length)


class PWArgs:
    """
    Higher-level bridge between Python values and input values.
    """

    def __init__(self, obj):
        self._obj = obj

    def __del__(self):
        if self._obj is None:
            return
        free(self._obj)
        self._obj = None

    @classmethod
    def from_python(cls, value, **kwargs):
        """ Convert a Python value to a PWArgs.
        """
        return cls(create(value, **kwargs))

    def __repr__(self):
        return "<{} obj={!r}>".format(self.__class__.__name__, self._obj)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019, 2020, Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import os
import subprocess
import sys

from distutils.util import get_platform

try:
    from setuptools import Extension, setup
    from setuptools.command.build_ext import build_ext
except ImportError:
    from distutils.core import Extension, setup
    from distutils.command.build_ext import build_ext

metadata = {}
with open("sq_native/__about__.py", "r") as fp:
    exec(fp.read(), metadata)


# inspired by https://github.com/m-pilia/disptools/blob/d96186cfd5dc5b3a73ec785b0dc443d6e4d9948a/setup.py#L37
class CMakeExtension(Extension):

    def __init__(self, dest_module, cmakelists_dir=".", target=None, options=None, sources=[], **kwa):
        Extension.__init__(self, dest_module, sources=sources, **kwa)
        self.cmakelists_dir = os.path.abspath(cmakelists_dir)
        self.target = target
        self.options = options


class CMakeBuild(build_ext):

    def get_ext_filename(self, name):
        if os.name == "posix" and sys.platform == "darwin":
            prefix, ext = "lib", ".dylib"
        elif sys.platform == "win32":
            prefix, ext = "", ".dll"
        else:
            prefix, ext = "lib", ".so"
        parts = name.split(".")
        last = prefix + parts.pop(-1) + ext
        return os.path.join(*(parts + [last]))

    def build_extensions(self):
        cmake_bin = os.getenv("CMAKE_BIN", "cmake")
        try:
            subprocess.check_output([cmake_bin, "--version"])
        except OSError:
            raise RuntimeError("Cannot find CMake executable")

        for ext in self.extensions:
            extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
            cfg = os.getenv("SQ_NATIVE_BUILD_MODE", "Release")

            cmake_args = [
                "-DCMAKE_BUILD_TYPE=%s" % cfg,
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir),
                "-DCMAKE_RUNTIME_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir),
                "-DCMAKE_ARCHIVE_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), self.build_temp),
            ]

            if ext.options is not None:
                cmake_args.extend(ext.options)

            if not os.path.exists(self.build_temp):
                os.makedirs(self.build_temp)

            # Config and build the extension
            subprocess.check_call([cmake_bin, ext.cmakelists_dir] + cmake_args,
                                  cwd=self.build_temp)
            cmd = [cmake_bin, "--build", ".", "--config", cfg]
            if ext.target is not None:
                cmd.extend(["--target", ext.target])
            subprocess.check_call(cmd, cwd=self.build_temp)


ext_modules = []
root_dir = os.path.dirname(os.path.abspath(__file__))
waf_src = os.path.join(root_dir, "PowerWAF")
if os.path.exists(waf_src):
    options = ["-DKEEP_SYMBOL_FILE=FALSE"]
    if get_platform() == "win-amd64":
        options.extend(["-A", "X64"])
    elif os.path.exists("/etc/alpine-release"):
        options.extend(["-DLIBC=muslc"])
    ext_modules.append(
        CMakeExtension("sq_native.Sqreen", "PowerWAF", "Sqreen", options))


setup(
    name="sq-native",
    description="Native module for the Python agent.",
    version=metadata["__version__"],
    author=metadata["__author__"],
    author_email=metadata["__email__"],
    license="proprietary",
    packages=["sq_native"],
    ext_modules=ext_modules,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: Other/Proprietary License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    cmdclass={"build_ext": CMakeBuild},
    test_suite="tests",
    tests_require=["pytest", "mock"],
)
